/*!\file window.c
 *
 * \brief sdlGLexemple05-1.2 : exemple de création de sphère ombrée et
 * manipulation des lumières.
 *
 * \author Farès BELHADJ (Lire COPYING) amsi@ai.univ-paris8.fr 
 *
 * \date October 18, 2013
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

static SDL_Surface * sdlInit(int w, int h);
static void          sdlInitGL(SDL_Surface * s);
static void          sdlResizeGL(SDL_Surface * s);
static void          sdlLoop(SDL_Surface * s);
static int           sdlManageEvents(SDL_Surface * s);
static void          sdlDraw(SDL_Surface * s);
static void          sdlSphere(int);
static inline void   sdlFPS(void);

static int longiMax = 40, latiMax = 40, pause = 0;
/*!\brief la lumière est positionnelle si cette variable est à 1, elle
 * est directionnelle sinon */
static int posi = 1;
/*!\brief atténuation de la lumière positionnelle, ici cette variable
   est utilisée comme coefficient d'atténuation quadratique */
static GLfloat lightAttenuation = 0.0;


/*!\brief initialise la surface d'affichage puis appelle la boucle
 * principale de gestioon d'évènement/dessin */
int main(int argc, char ** argv) {
  SDL_Surface * s;
  s = sdlInit(640, 480);
  sdlLoop(s);
  sdlSphere(1);
  return 0;
}

/*!\brief initialisation de SDL pour une utilisation avec OpenGL.
 */
static SDL_Surface * sdlInit(int w, int h) {
  SDL_Surface * s;
  const SDL_VideoInfo * vi = NULL;
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Impossible d'initialiser SDL: %s\n", SDL_GetError());
    exit(1);
  }
  if( (vi = SDL_GetVideoInfo()) == 0 ) {
    fprintf(stderr, "Echec de SDL_GetVideoInfo(): %s\n", SDL_GetError());
    exit(1);
  }
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE,     8 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE,   8 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE,    8 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE,  16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  if ( (s = SDL_SetVideoMode(w, h, vi->vfmt->BitsPerPixel,
			     SDL_OPENGL | SDL_RESIZABLE)) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le mode video : %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);
  sdlInitGL(s);
  return s;
}

/*!\brief configurer les options GL.*/
static void sdlInitGL(SDL_Surface * s) {
  glShadeModel(GL_SMOOTH);
  glDepthFunc(GL_LEQUAL);
  glFrontFace(GL_CCW);
  glCullFace(GL_BACK);
  glClearColor(0, 0, 0, 0);
  glClearDepth(1.0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glDisable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_NORMALIZE);

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  /* activation du lighting (ombrage) et de la lumière 0 et 1 */
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHT1);
  
  sdlResizeGL(s);
}

/*!\brief CallBack en cas de resize : modifie la matrice de projection en
 * fonction du rapport w / h (liés à la surface \a s)
 */
static void sdlResizeGL(SDL_Surface * s) {
  glViewport(0, 0, s->w, s->h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, s->w / (float)s->h, 0.1, 100.0);
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
}

/*!\brief loop principale : sdlManageEvents renvoie vrai si la fenetre est
 * inactive => tantque inactive dors une milliseconde.
 * sinon dessiner puis afficher le framerate
 */
static void sdlLoop(SDL_Surface * s) {
  for(;;) {
    while(sdlManageEvents(s)) SDL_Delay(1);
    sdlDraw(s);
    sdlFPS();
  }
}

/*!\brief gere les evenements ; aussi renvoie vrai si la
 * fenetre est inactive, sinon faux.
 *
 * Les evenements sont : 
 *
 * escape, q, Q : quitter
 *
 * F1 : fullscreen On / Off
 *
 * <- & -> : diminuer et augmenter le nombre de tranches verticales
 *
 * flêche bas et haut : diminuer et augmenter le nombre de tranches horizontales
 *
 * p / espace : pause
 *
 * w : mode wire frame
 *
 * f : mode flat / smooth
 *
 * l : mode lumière
 * 
 * d : lumière directionnelle/positionnelle
 * 
 * a/z : augmentation / dimunition de l'atténuation
 * 
 * redimensionnement de fenêtre : appeler \ref sdlResizeGL
 *
 * \return vrai si la fenêtre est inactive, faux sinon.
 */
static int sdlManageEvents(SDL_Surface * s) {
  GLint v[2];
  static int active = 1;
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case SDLK_UP:
	latiMax++;
	goto sphereParamsChanged;
      case SDLK_DOWN:
	latiMax--;
	goto sphereParamsChanged;
      case SDLK_RIGHT:
	longiMax++;
	goto sphereParamsChanged;
      case SDLK_LEFT:
	longiMax--;
	goto sphereParamsChanged;
      case 'a':
	lightAttenuation += 0.1;
	break;
      case 'z':
	lightAttenuation -= 0.1;
	break;
      case 'd':
	posi = !posi;
	break;
      case ' ':
      case 'p':
	pause = !pause;
	break;
      case 'w':
	glGetIntegerv(GL_POLYGON_MODE, v);
	if(v[0] == GL_FILL)
	  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	else
	  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	break;
      case 'f':
	glGetIntegerv(GL_SHADE_MODEL, v);
	if(v[0] == GL_FLAT)
	  glShadeModel(GL_SMOOTH);
	else
	  glShadeModel(GL_FLAT);
	break;
      case 'l':
	if(glIsEnabled(GL_LIGHTING)) {
	  glDisable(GL_LIGHTING);
	  glDisable(GL_LIGHT0);
	  glDisable(GL_LIGHT1);
	} else {
	  glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0);
	  glEnable(GL_LIGHT1);
	}
	break;
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      case SDLK_F1:
	SDL_WM_ToggleFullScreen(s);
	break;
      default:
	fprintf(stderr, "La touche %s a ete pressee\n",
		SDL_GetKeyName(event.key.keysym.sym));
	break;
      }
      break;
    case SDL_KEYUP:
      break;
    case SDL_ACTIVEEVENT:
      if(event.active.state & SDL_APPACTIVE)
	active = event.active.gain;
      break;
    case SDL_VIDEORESIZE:
      if( (s = SDL_SetVideoMode(event.resize.w, event.resize.h, 
				s->format->BitsPerPixel, s->flags)) == NULL ) {
	fprintf(stderr, "Echec de SDL_SetVideoMode apres le resize : %s\n", 
		SDL_GetError());
	exit(1);
      }
      sdlResizeGL(s);
      break;
    case SDL_QUIT:
      exit(0);
    }
  return !active;
 sphereParamsChanged:
  if(longiMax < 3)
    longiMax = 3;
  else if(latiMax < 3)
    latiMax = 3;
  else
    sdlSphere(1);
  return !active;
}

/*!\brief envoie la scène à dessiner, ici nous sommes dans l'espace
 * MODELVIEW.
 */
static void sdlDraw(SDL_Surface * s) {
  static int t0 = 0, t1;
  static GLfloat a = 0.0;
  GLfloat dt;
  /*!\brief couleur de la composante ambiante de la lumière 0 */
  GLfloat light_ambient0[]  = {  0.0, 0.0, 0.4, 1.0 }; 
  /* couleur de la composante diffuse  de la lumière 0 */
  GLfloat light_diffuse0[]  = {  1.0, 0.0, 0.0, 1.0 }; 
  /* position/direction de la lumière 0, si le w (4e comp) est à 1.0,
   * la lumière est positionnelle, sinon elle est directionnelle */
  GLfloat light_position0[] = { -1.2, 1.2, 1.2, 1.0 }; 
  /*!\brief couleur de la composante ambiante de la lumière 1 */
  GLfloat light_ambient1[]  = {  0.0, 0.0, 0.0, 1.0 };
  /* couleur de la composante diffuse  de la lumière 1 */
  GLfloat light_diffuse1[]  = {  0.0, 1.0, 0.0, 1.0 };
  /* position/direction de la lumière 1, si le w (4e comp) est à 1.0,
   * la lumière est positionnelle, sinon elle est directionnelle */
  GLfloat light_position1[] = {  1.2, 1.2, 1.2, 1.0 };

  /* La lumière est positionnelle (1.0 en w) si posi est vrai */
  light_position0[3] = posi ? 1.0 : 0.0;
  light_position1[3] = posi ? 1.0 : 0.0;

  /* récupération du dt */
  t1 = SDL_GetTicks();
  dt = (t1 - t0) / 1000.0;
  t0 = t1;
  /* fin de récupération du dt */

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  glTranslatef(0.0f,0.0f,-3.5f); 

  glPushMatrix();
  glRotatef(2 * a, 0, 0, 1);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position0);
  glPopMatrix();

  glPushMatrix();
  glRotatef(-3 * a, 0, 1, 0);
  glLightfv(GL_LIGHT1, GL_POSITION, light_position1);
  glPopMatrix();

  glLightfv (GL_LIGHT0, GL_DIFFUSE,  light_diffuse0);
  glLightfv (GL_LIGHT0, GL_AMBIENT,  light_ambient0);
  glLightf  (GL_LIGHT0, GL_QUADRATIC_ATTENUATION, lightAttenuation);
  glLightfv (GL_LIGHT1, GL_AMBIENT,  light_ambient1);
  glLightfv (GL_LIGHT1, GL_DIFFUSE,  light_diffuse1);
  if(!pause)
    a += 90 * dt;
  /* regarder d'un peu plus haut */
  glRotatef(30, 1, 0, 0);
  glRotatef(a, 0, 1, 0);
  sdlSphere(0);
  SDL_GL_SwapBuffers();
}

/*\!brief dessine une sphère unitaire coloriée ou ombrée (vecteurs
 * normaux) ; cette fonction utilise des QUADS.
 *
 * \param onlyFreeData demande à la fonction de libérer sa donnée sans
 * rien émettre à GL.
 */
static void sdlSphere(int onlyFreeData) {
  static GLfloat * vertices = NULL;
  int i, e;
  GLfloat phi0, phi1, theta0, theta1;  
  if(onlyFreeData) {
    if(vertices) {
      free(vertices);
      vertices = NULL;
    }
    return;
  }
  if(!vertices) {
    vertices = malloc(2 * latiMax * 2 * longiMax * 3 * sizeof *vertices);
    assert(vertices);
  }
  for(e = 0 ; e < latiMax; e++) {
    theta0 = -(M_PI / 2) + (e * M_PI) / (GLfloat)(latiMax);
    theta1 = -(M_PI / 2) + ((e + 1) * M_PI) / (GLfloat)(latiMax);
    for(i = 0; i < longiMax; i++) {
      phi0 = (2.0 * M_PI * i) / (GLfloat)longiMax;
      phi1 = (2.0 * M_PI * (i + 1)) / (GLfloat)longiMax;
      vertices[12 * (longiMax * e + i) + 0] = cos(theta0) * cos(phi0);
      vertices[12 * (longiMax * e + i) + 1] = sin(theta0);
      vertices[12 * (longiMax * e + i) + 2] = cos(theta0) * sin(phi0);
      vertices[12 * (longiMax * e + i) + 3] = cos(theta0) * cos(phi1);
      vertices[12 * (longiMax * e + i) + 4] = sin(theta0);
      vertices[12 * (longiMax * e + i) + 5] = cos(theta0) * sin(phi1);
      vertices[12 * (longiMax * e + i) + 6] = cos(theta1) * cos(phi1);
      vertices[12 * (longiMax * e + i) + 7] = sin(theta1);
      vertices[12 * (longiMax * e + i) + 8] = cos(theta1) * sin(phi1);
      vertices[12 * (longiMax * e + i) + 9] = cos(theta1) * cos(phi0);
      vertices[12 * (longiMax * e + i) + 10] = sin(theta1);
      vertices[12 * (longiMax * e + i) + 11] = cos(theta1) * sin(phi0);
    }
  }
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);
  glVertexPointer(3, GL_FLOAT, 0, vertices);
  glNormalPointer(GL_FLOAT, 0, vertices);
  glColorPointer(3, GL_FLOAT, 0, vertices);
  glDrawArrays(GL_QUADS, 0, 2 * 2 * longiMax * latiMax);
  glDisableClientState(GL_COLOR_ARRAY);
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
}

/*!\brief compter le nombre de frames affiches. imprimer le
 * framerate toutes les 5 secondes.
 */
static inline void sdlFPS(void) {
  Uint32 t;
  static Uint32 t0 = 0, f = 0;
  f++;
  t = SDL_GetTicks();
  if(t - t0 > 5000) {
    fprintf(stderr, "sdlFPS:%6d\n",
	    (int)(1000.0 * f / (t - t0)));
    t0 = t;
    f  = 0;
  }
}

